module edetect_io_box
  
  use errorHandling
  use param  
  use dal

  implicit none

  private

  public cp_head, get_files, co_image, co_bgima, co_expima, co_mask
  
! MAX_IMA              max number of images
! MAX_INS              max number of different instruments 
! MAX_BANDS            max number of energy bands

  integer, parameter, public :: max_ima = 240
  integer, parameter, public :: max_ins = 60
  integer, parameter, public :: max_bands = 6

! NUM_INSTR            number of instruments
! NUM_IMAGE(MAX_INS)   number of images per instrumet
! NIM                  total number of images

  integer, public :: num_instr, nim  
  integer, dimension(max_ins), public :: num_image       
  integer, dimension(max_ima), public :: pimin, pimax

! IMAGE                input images
! BG_IMAGE             background images
! EXP_IMAGE            exposure images

  integer(kind=int16), dimension(:,:,:), allocatable, public  :: image
  real, dimension (:,:,:), allocatable, public :: exp_image, bg_image
  logical, dimension(:,:,:), allocatable, public :: mask_image

  integer, dimension(2), public :: npix

! IP(MAX_INS,MAX_BANDS)pointer field to manage images

  integer, dimension(max_ins, max_bands), public :: ip
  integer, dimension(max_ins), public :: expip

  integer, dimension(max_ima) :: ixb0, iyb0
  integer :: imbuff

  real(kind=single), dimension(max_ima), public :: ecf

  real(kind=single), public :: xcen1, ycen1
  real(kind=double), public :: ra_pnt, dec_pnt, pa_pnt
  real(kind=double), public :: crpix1,crpix2,crval1,crval2, &
                               crdelt1,crdelt2,crota2
  real(kind=single), dimension(3), public :: xoff, yoff
  character(len=64), public :: submode

  real(kind=double) :: pi = 3.141592654

  public ixb, iyb, imbuff

contains

  !---------------------------------------------------------------------
  integer function ixb(ix,im,ncom)
  !---------------------------------------------------------------------

  integer :: ix, im, ncom

  ixb = max(min(ix-ixb0(im)/ncom,imbuff),1)

  end function ixb

  !---------------------------------------------------------------------
  integer function iyb(iy,im,ncom)
  !---------------------------------------------------------------------

  integer :: iy, im, ncom

  iyb = max(min(iy-iyb0(im)/ncom,imbuff),1)

  end function iyb



  !---------------------------------------------------------------------
  subroutine cp_attr (set1, set2, arr, attr)
  !---------------------------------------------------------------------
    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT), intent(IN) :: arr
    character(len=*), dimension(:), intent(IN) :: attr

    integer i

    do i=1,size(attr)
      if (hasAttribute(set1,attr(i))) then
        call setAttribute(set2,attribute(set1,attr(i)))
      else 
        if (hasAttribute(arr,attr(i))) then
          call setAttribute(set2,attribute(arr,attr(i)))
        end if
      end if
    end do

  end subroutine cp_attr

  !---------------------------------------------------------------------
  subroutine cp_head (set1, set2, content, mod)
  !---------------------------------------------------------------------

    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT) :: arr
    character(len=*), intent (IN) :: content
    integer, intent(IN) :: mod
    character(len=2), dimension(3) :: strInst
    character(len=8)  :: strEmin,strEmax 
    character(len=1)  :: strBand 
    character(len=80) :: strComment
    integer i,j,ipoint    

    data strInst /"PN","M1","M2"/

    arr = array (set1,"PRIMARY",read)

    if (mod == 1 .or. mod == 2 .or. mod == 3) then
      call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
    end if

    if (mod == 2) then
      call cp_attr(set1,set2,arr,(/"BAND"/) )
    end if

    !if (mod == 3) then
    !  call cp_attr(set1,set2,arr,(/"INSTRUME","DETNAME "/) )
    !end if
    
    if (mod < 3) then  
      call cp_attr(set1,set2,arr, &
      (/ "TELESCOP","DATAMODE","FILTER  ","OBS_MODE","OBS_ID  ","EXP_ID  ", &
         "EXPIDSTR","ORIGIN  ","OBJECT  ","RA_PNT  ","DEC_PNT ","PA_PNT  ", &
         "CRPIX1  ","CRPIX2  ","CRVAL1  ","CRVAL2  ","CDELT1  ", &
         "CDELT2  ","CTYPE1  ","CTYPE2  ","SUBMODE ","RADECSYS","EQUINOX ", &
         "EXPOSURE","DATE-OBS","DATE-END","OBSERVER","ATT_SRC "/) )
    else
       !for tables, rename WCS keywords:
       call cp_attr(set1,set2,arr, &
         (/ "TELESCOP","DATAMODE","FILTER  ","OBS_MODE","OBS_ID  ","EXP_ID  ", &
            "EXPIDSTR","ORIGIN  ","OBJECT  ","RA_PNT  ","DEC_PNT ","PA_PNT  ", &
            "RADECSYS","EQUINOX ","EXPOSURE","DATE-OBS","DATE-END","OBSERVER", &
            "ATT_SRC ","SUBMODE " /) )
       call cp_wcs2table(set1,set2,"real32",arr, &
          (/ "CRPIX1  ","CRPIX2  ","CRVAL1  ","CRVAL2  ","CDELT1  ", &
             "CDELT2  "/) )
       call cp_wcs2table(set1,set2,"string",arr, (/ "CTYPE1  ","CTYPE2  "/) )
     end if

    do i=1,num_instr
      do j=1,num_image(i)
        ipoint=ip(i,j)
        if (ipoint > 0) then
          write(strBand,'(I1)') j 
          strEmin = strInst(expip(i))//"_"//strBand//"_ELO"
          strEmax = strInst(expip(i))//"_"//strBand//"_EHI"
          strComment = "E_min for Instr "//strInst(expip(i))//" , band "//strBand   
          call setAttribute(set2,strEmin,pimin(ipoint),"PI chan",strComment)
          strComment = "E_min for Instr "//strInst(expip(i))//" , band "//strBand   
          call setAttribute(set2,strEmax,pimax(ipoint),"PI chan",strComment)
        endif
      end do
    end do


    call setAttribute(set2,"CONTENT", content, "Content of file")

    call release(arr)

  end subroutine cp_head

  !---------------------------------------------------------------------
  subroutine cp_wcs2table (set1, set2, dtype, arr, attr)
  !---------------------------------------------------------------------
    type(DataSetT), intent(IN) :: set1, set2
    type(ArrayT), intent(IN) :: arr
    character(len=*), dimension(:), intent(IN) :: attr
    character(len=*), intent(IN) :: dtype
    character(len=8) attname
    real value
    character(len=80) svalue,comment 
    integer i

    do i=1,size(attr)
      comment="copy of WCS keyword "//attr(i)//" from input image" 
      if (dtype == "real32") then 
        if (hasAttribute(set1,attr(i))) then
          value=real32Attribute(set1,attr(i))
          attname = "T"//attr(i)
          call setAttribute(set2,attname,value,"",trim(comment))
        else 
          if (hasAttribute(arr,attr(i))) then
            value=real32Attribute(arr,attr(i))
            attname = "T"//attr(i)
            call setAttribute(set2,attname,value,"",trim(comment))
          end if
        end if
      elseif (dtype == "string") then
        if (hasAttribute(set1,attr(i))) then
          svalue=stringAttribute(set1,attr(i))
          attname = "T"//attr(i)
          call setAttribute(set2,attname,trim(svalue),trim(comment))
        else 
          if (hasAttribute(arr,attr(i))) then
            svalue=stringAttribute(arr,attr(i))
            attname = "T"//attr(i)
            call setAttribute(set2,attname,trim(svalue),trim(comment))
          end if
        end if 
      end if 
    end do

  end subroutine cp_wcs2table

  !--------------------------------------------------------------------
  subroutine get_files (param_name, imag, instrume, band, first)
  !---------------------------------------------------------------------
 
    character(len=*), intent(IN) :: param_name
    logical, intent(IN) :: first
    character(len=1024), dimension(*), intent(OUT) :: imag
!    character(len=16), dimension(*), intent(OUT) :: instrume 
    character(len=21), dimension(*), intent(OUT) :: instrume 
    integer, dimension(*), intent(OUT) :: band
    character(len=80) errMess

    integer :: i, j, ib
    integer :: nimage, ipoint
!    character(len=7) :: teles, instru, expid  
    character(len=3) :: teles
    character(len=5) :: instru
!   character(len=11) :: expid
    character(len=13) :: expid
!    character(len=16),dimension(max_ima) :: ins 
    character(len=21),dimension(max_ima) :: ins 
 
    logical, dimension(max_ima) :: double

    type(DataSetT) :: set
    type(ArrayT) :: arr

    nimage = parameterCount(param_name) 
    if (nimage == 0) then
      call error("MissingParameter", "Missing input file name")
    end if
 
    if (first) then 
    
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)


        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"EXP_ID")) then
          expid = stringAttribute(set,"EXP_ID")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"EXP_ID")) then
            expid = stringAttribute(arr,"EXP_ID")
          else
            expid = "---"
            call warning("MissingAttribute","Keyword EXP_ID is missing")
          end if
          call release(arr)
        end if
!        write(*,*) "instrume:", trim(teles) , trim(instru) , trim(expid) 
        instrume(i) = trim(teles) // trim(instru) // trim(expid)
 
        if (hasAttribute(set,"BAND")) then
          band(i) = int32Attribute(set, "BAND")
        else
          band(i) = 7
        end if
        call release(set)
      end do

      num_instr = 0
      double = .false.    
      ip = -1
      do i=1,nimage        
        if (.not. double(i)) num_instr = num_instr+1
        ib = 1 
        do while (ip(num_instr,ib) /= -1)
          ib = ib+1
        end do
        ip(num_instr,ib) = i
        num_image(num_instr) = ib
        do j=i+1,nimage
          if (instrume(i) == instrume(j)) double(j) = .true.
        end do
      end do

      do i=1,num_instr
        ipoint=ip(i,1) 
        if (index(instrume(ipoint),"PN") > 0 .or. &
            index(instrume(ipoint),"pn") > 0) then 
          expip(i) = 1
        else if (index(instrume(ipoint),"MOS1") > 0 .or. &
                 index(instrume(ipoint),"mos1") > 0) then
          expip(i) = 2
        else if (index(instrume(ipoint),"MOS2") > 0 .or. &
                 index(instrume(ipoint),"mos2") > 0) then
          expip(i) = 3
        else
          expip(i) = 4
          call error("WrongInst","Unknown instrument")
        end if
      end do
        
    else

      if (param_name(1:3) == "det") then
        if (nimage /= num_instr) then
           call error("FileMismatch", &
             "# detector masks /= # instruments")
        end if
      else 
        if (nimage /= nim) then
          call error("FileMismatch","Inconsistent number of input images")
        end if
      end if
 
      do i=1,nimage
        imag(i) = stringParameter(param_name,i-1)
        set = dataSet(imag(i), READ)

        if (hasAttribute(set,"TELESCOP")) then
          teles = stringAttribute(set,"TELESCOP")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"TELESCOP")) then
            teles = stringAttribute(arr,"TELESCOP")
          else
            teles = "---"
            call warning("MissingAttribute","Keyword TELESCOP is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"INSTRUME")) then
          instru = stringAttribute(set,"INSTRUME")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"INSTRUME")) then
            instru = stringAttribute(arr,"INSTRUME")
          else
            instru = "---"
            call warning("MissingAttribute","Keyword INSTRUME is missing")
          end if
          call release(arr)
        end if

        if (hasAttribute(set,"EXP_ID")) then
          expid = stringAttribute(set,"EXP_ID")
        else 
          arr = array(set,"PRIMARY",read) 
          if (hasAttribute(arr,"EXP_ID")) then
            expid = stringAttribute(arr,"EXP_ID")
          else
            expid = "---"
            call warning("MissingAttribute","Keyword EXP_ID is missing")
          end if
          call release(arr)
        end if
        ins(i) = trim(teles) // trim(instru) // trim(expid)


        call release(set)

      end do


      if (param_name(1:3) == "det") then 
         do i=1,num_instr
           do j=1,num_image(i)
             ipoint=ip(i,j)
             if (ins(i) /= instrume(ipoint)) then 
!               write(*,*) i, ipoint, ins(i), instrume(ipoint)
               errMess = ": Wrong detector mask" 
               call error ("FileMismatch",errMess) 
             end if
           end do
         end do
      end if 

    end if

    nim = ip(num_instr,num_image(num_instr))

    

  end subroutine get_files

  !---------------------------------------------------------------------
  subroutine co_image (fname, start, step, npix, exposure, xcen, ycen, &
                       ra_pnt, dec_pnt, pa_pnt, &
                       crpix1, crpix2, crval1, crval2, &
                       crdelt1,crdelt2,crota2,xoff,yoff,inst,ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix
    real(kind=double), dimension(2), intent(out) :: start, step 
    real(kind=double), intent(out) :: ra_pnt, dec_pnt, pa_pnt, &
                       crpix1, crpix2, crval1, crval2, &
                       crdelt1, crdelt2, crota2 
    real(kind=double) ::cos_pa, sin_pa, rr, dr, r0r, d0r, &
                       dra, ddec, xr, yr, dist1, dist2 
    real(kind=single) :: exposure, xcen, ycen, r4huge
    real(kind=double) :: r8huge
    integer, dimension(3) :: ixoff, iyoff
    integer :: i, ix, iy, iix, iiy, atype, itot, i4huge
    integer :: ix0, ix1, iy0, iy1

    integer, dimension(2), save :: npix_save
    logical, save :: first = .true.
    logical :: empty

    integer(kind=int32), dimension(:,:), pointer :: i4arr
    integer(kind=int16), dimension(:,:), pointer :: i2arr
    integer(kind=int8), dimension(:,:), pointer :: i1arr
    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=double), dimension(:,:), pointer :: r8arr
    integer, dimension(:), pointer :: shape

    integer(kind=int32), dimension(:), allocatable :: ipx, ipy
    integer(kind=int16) :: i2

    type(datasett) :: set
    type(arrayt) :: arr
    !!--JR
    logical :: withimbuff
    !!-- JR ------
    withimbuff = booleanParameter("withimagebuffersize")
    !!------------
    !!-- JR ------
    if (withimbuff) then
      imbuff = intParameter("imagebuffersize")
      withimbuff = .true.
      else
      imbuff = 640
      withimbuff = .false.
    endif
    !!------------

    i4huge=huge(i2)
    r4huge=huge(i2)
    r8huge=huge(i2)

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)

    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    !! -- JR -----------------------------
!!!    if (withimbuff) then
!!!      imbuff = imbuff
!!!      else
    if (.not. withimbuff) then
      imbuff = max(npix(1),npix(2))
    end if
    !! -----------------------------------

    if (first) then
      npix_save=npix
      allocate (image(nim,imbuff,imbuff))
      first=.false.
    end if

    allocate (ipx(0:npix(1)-1))
    allocate (ipy(0:npix(2)-1))

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    itot=0
    atype = arrayDataType(arr)
     if (atype == Integer8) then
       i1arr => int8array2data(arr)

!-----------------------------------------------------------------------
!------- (JR) ----------------------------------------------------------
       ix0=1
       iy0=1
       ix1=1
       iy1=1
       ipx=sum(i1arr,2)
       ipy=sum(i1arr,1)

       empty = .true.
       do iix=0,npix(1)-1
         if (empty .and. ipx(iix)>0) then
           ix0=iix
           empty = .false.
         end if
         if (ipx(iix)>0) ix1=iix
       end do
       empty = .true.
       do iiy=0,npix(2)-1
         if (empty .and. ipy(iiy)>0) then
           iy0=iiy
           empty = .false.
         end if
         if (ipy(iiy)>0) iy1=iiy
       end do
       ixb0(ipoint) = (max(ix0,1)/16) * 16
       iyb0(ipoint) = (max(iy0,1)/16) * 16
       if (ix1-ixb0(ipoint)>imbuff .or. iy1-iyb0(ipoint)>imbuff ) &
          call error("SizeMissmatch","Increase ImageBufferSize")
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------


       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) then
!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
             if (i1arr(iix,iiy)>0) then
             image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = i1arr(iix,iiy)
             itot=itot+i1arr(iix,iiy)
             end if
            end if
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
         end do
       end do 
     else if (atype == Integer16) then
       i2arr => int16array2data(arr)

!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
       ix0=1
       iy0=1
       ix1=1
       iy1=1
       ipx=sum(i2arr,2)
       ipy=sum(i2arr,1)

       empty = .true.
       do iix=0,npix(1)-1
         if (empty .and. ipx(iix)>0) then
           ix0=iix
           empty = .false.
         end if
         if (ipx(iix)>0) ix1=iix
       end do
       empty = .true.
       do iiy=0,npix(2)-1
         if (empty .and. ipy(iiy)>0) then
           iy0=iiy
           empty = .false.
         end if
         if (ipy(iiy)>0) iy1=iiy
       end do
       ixb0(ipoint) = (max(ix0,1)/16) * 16
       iyb0(ipoint) = (max(iy0,1)/16) * 16
       if (ix1-ixb0(ipoint)>imbuff .or. iy1-iyb0(ipoint)>imbuff ) &
          call error("SizeMissmatch","Increase ImageBufferSize")
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------


       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) then
!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
             if (i2arr(iix,iiy)>0) then
               image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = i2arr(iix,iiy)
               itot=itot+i2arr(iix,iiy)
             end if
            end if
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
         end do
       end do 
     else if (atype == Integer32) then 
       i4arr => int32array2data(arr)


!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
       ix0=1
       iy0=1
       ix1=1
       iy1=1
       ipx=sum(i4arr,2)
       ipy=sum(i4arr,1)

       empty = .true.
       do iix=0,npix(1)-1
         if (empty .and. ipx(iix)>0) then
           ix0=iix
           empty = .false.
         end if
         if (ipx(iix)>0) ix1=iix
       end do
       empty = .true.
       do iiy=0,npix(2)-1
         if (empty .and. ipy(iiy)>0) then
           iy0=iiy
           empty = .false.
         end if
         if (ipy(iiy)>0) iy1=iiy
       end do
       ixb0(ipoint) = (max(ix0,1)/16) * 16
       iyb0(ipoint) = (max(iy0,1)/16) * 16
       if (ix1-ixb0(ipoint)>imbuff .or. iy1-iyb0(ipoint)>imbuff ) &
          call error("SizeMissmatch","Increase ImageBufferSize")
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------

       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) then
!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
             if (i4arr(iix,iiy)>0) then
               !print *, ix,iy,ixb0(ipoint),iyb0(ipoint),ixb(ix,ipoint,1),iyb(iy,ipoint,1)
               image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = min(i4arr(iix,iiy),i4huge)
               itot=itot+i4arr(iix,iiy)
             end if
            end if
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
         end do
       end do 
     else if (atype == Real32) then
       r4arr => Real32array2data(arr)

!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
       ix0=1
       iy0=1
       ix1=1
       iy1=1
       ipx=sum(r4arr,2)
       ipy=sum(r4arr,1)

       empty = .true.
       do iix=0,npix(1)-1
         if (empty .and. ipx(iix)>0) then
           ix0=iix
           empty = .false.
         end if
         if (ipx(iix)>0) ix1=iix
       end do
       empty = .true.
       do iiy=0,npix(2)-1
         if (empty .and. ipy(iiy)>0) then
           iy0=iiy
           empty = .false.
         end if
         if (ipy(iiy)>0) iy1=iiy
       end do
       ixb0(ipoint) = (max(ix0,1)/16) * 16
       iyb0(ipoint) = (max(iy0,1)/16) * 16
       if (ix1-ixb0(ipoint)>imbuff .or. iy1-iyb0(ipoint)>imbuff) then
          call error("SizeMissmatch","Increase ImageBufferSize")
          print *, ixb0(ipoint),iyb0(ipoint),ix1,iy1
       end if
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------

       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) then
!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
            if (r4arr(iix,iiy)>0) then
               !print *, ix,iy,ixb0(ipoint),iyb0(ipoint),ixb(ix,ipoint,1),iyb(iy,ipoint,1)
               image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = nint(min(r4arr(iix,iiy),r4huge))
               itot=itot+r4arr(iix,iiy)
             end if
            end if
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
         end do
       end do 
     else if (atype == Real64) then
       r8arr => Real64array2data(arr)

!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
       ix0=1
       iy0=1
       ix1=1
       iy1=1
       ipx=sum(r8arr,2)
       ipy=sum(r8arr,1)

       empty = .true.
       do iix=0,npix(1)-1
         if (empty .and. ipx(iix)>0) then
           ix0=iix
           empty = .false.
         end if
         if (ipx(iix)>0) ix1=iix
       end do
       empty = .true.
       do iiy=0,npix(2)-1
         if (empty .and. ipy(iiy)>0) then
           iy0=iiy
           empty = .false.
         end if
         if (ipy(iiy)>0) iy1=iiy
       end do
       ixb0(ipoint) = (max(ix0,1)/16) * 16
       iyb0(ipoint) = (max(iy0,1)/16) * 16
       if (ix1-ixb0(ipoint)>imbuff .or. iy1-iyb0(ipoint)>imbuff ) &
          call error("SizeMissmatch","Increase ImageBufferSize")
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------

       do iy = 1,npix(2)
         do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) then
!-----------------------------------------------------------------------
!------------- (JR) ----------------------------------------------------
            if (r8arr(iix,iiy)>0) then
               image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = nint(min(r8arr(iix,iiy),r8huge))
               itot=itot+r8arr(iix,iiy)
             end if
            end if
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
         end do
       end do 
     else
       call error("WrongType","Input image has wrong type")
     end if
     
! this is for testing ---------------------------------
!     do ix=1,npix(1)
!       do iy=1,npix(2)
!         image(ipoint,ix,iy)=0
!       end do
!     end do 
!     nev=npix(1)*npix(2)*0.2
!     ! if (ipoint==1) iseed=-123456789 
!     do i=1,nev
!       ix=ran(iseed)*npix(1)+1
!       iy=ran(iseed)*npix(2)+1
!       if (ix.ge.1.and.ix.le.npix(1).and. &
!           iy.ge.1.and.iy.le.npix(2)) &
!         image(ipoint,ix,iy)=image(ipoint,ix,iy)+1
!     end do
!------------------------------------------------------

     if (itot==0) call error("EmptyImage","Input image is empty")

    start(1) = 1d0
    start(2) = 1d0
    step(1) = 1d0
    step(2) = 1d0

    if (hasAttribute(set,"EXPOSURE")) then
      exposure = real32Attribute(set,"EXPOSURE")
    else if (hasAttribute(arr,"EXPOSURE")) then
      exposure = real32Attribute(arr,"EXPOSURE")
    else
      exposure = 1.0
      call warning("MissingAttribute", &
                   "No exposure keyword present: set to 1.0")
    end if
    
    if (hasAttribute(set,"SUBMODE")) then
      submode = stringAttribute(set,"SUBMODE")
    else if (hasAttribute(arr,"SUBMODE")) then
      submode = stringAttribute(arr,"SUBMODE")
    else 
      submode="PrimeFullWindow"
    end if 

    if (hasAttribute(set,"RA_PNT")) then
      ra_pnt = real64Attribute(set,"RA_PNT")
    else  
      if (hasAttribute(arr,"RA_PNT")) then
        ra_pnt = real64Attribute(arr,"RA_PNT")
      else
        ra_pnt = 0d0
        call warning("MissingAttribute","Keyword RA_PNT is missing")
      end if
    end if

    if (hasAttribute(set,"DEC_PNT")) then
      dec_pnt = real64Attribute(set,"DEC_PNT")
    else  
      if (hasAttribute(arr,"DEC_PNT")) then
        dec_pnt = real64Attribute(arr,"DEC_PNT")
      else
        dec_pnt = 0d0
        call warning("MissingAttribute","Keyword DEC_PNT is missing")
      end if
    end if

    if (hasAttribute(set,"PA_PNT")) then
      pa_pnt = real64Attribute(set,"PA_PNT")
    else  
      if (hasAttribute(arr,"PA_PNT")) then
        pa_pnt = real64Attribute(arr,"PA_PNT")
      else
        pa_pnt = 0d0
        call warning("MissingAttribute","Keyword PA_PNT is missing")
      end if
    end if

    if (hasAttribute(arr,"CRPIX1")) then
      crpix1 = real64Attribute(arr,"CRPIX1")
    else
      if (hasAttribute(set,"CRPIX1")) then
        crpix1 = real64Attribute(set,"CRPIX1")
      else
        crpix1 = 0d0
        call warning("MissingAttribute","Keyword CRPIX1 is missing")
      end if
    end if

    if (hasAttribute(arr,"CRPIX2")) then
      crpix2 = real64Attribute(arr,"CRPIX2")
    else
      if (hasAttribute(set,"CRPIX2")) then
        crpix2 = real64Attribute(set,"CRPIX2")
      else
        crpix2 = 0d0
        call warning("MissingAttribute","Keyword CRPIX2 is missing")
      end if 
    end if

    if (hasAttribute(arr,"CRVAL1")) then
      crval1 = real64Attribute(arr,"CRVAL1")
    else
      if (hasAttribute(set,"CRVAL1")) then
        crval1 = real64Attribute(set,"CRVAL1")
      else
        crval1 = 0d0
        call warning("MissingAttribute","Keyword CRVAL1 is missing")
      end if
    end if

    if (hasAttribute(arr,"CRVAL2")) then
      crval2 = real64Attribute(arr,"CRVAL2")
    else
      if (hasAttribute(set,"CRVAL2")) then
        crval2 = real64Attribute(set,"CRVAL2")
      else
        crval2 = 0d0
        call warning("MissingAttribute","Keyword CRVAL2 is missing")
      end if
    end if

    if (hasAttribute(arr,"CDELT1")) then
      crdelt1 = real64Attribute(arr,"CDELT1")
    else
      if (hasAttribute(set,"CDELT1")) then
        crdelt1 = real64Attribute(set,"CDELT1")
      else
        crdelt1 = 0d0
        call warning("MissingAttribute","Keyword CRDELT1 is missing")
      end if
    end if

    if (hasAttribute(arr,"CDELT2")) then
      crdelt2 = real64Attribute(arr,"CDELT2")
    else
      if (hasAttribute(set,"CDELT2")) then
        crdelt2 = real64Attribute(set,"CDELT2")
      else
        crdelt2 = 0d0
        call warning("MissingAttribute","Keyword CDELT2 is missing")
      end if
    end if
 
    if (hasAttribute(arr,"CROTA2")) then
       crota2 = real64Attribute(arr,"CROTA2")
    else
      if (hasAttribute(set,"CROTA2")) then
        crota2 = real64Attribute(set,"CROTA2")
      else
        crota2 = 0d0
!        call warning("MissingAttribute","Keyword CROTA2 is missing")
      end if
    end if

    if (crdelt1 > 0.0 .and. crdelt2 > 0.0) then
      dist1=0.36d0/abs(crdelt1)
      dist2=0.36d0/abs(crdelt2)
    else
      dist1=320.0d0
      dist2=320.0d0
    end if
    if (crpix1 < -dist1 .or. crpix1 > npix(1)+dist2 .or. &
        crpix2 < -dist2 .or. crpix2 > npix(2)+dist2) then
      call error("WrongRefPix","Reference pixel is outside of FOV")
    end if

    cos_pa = dcos(crota2 * pi/180d0)
    sin_pa = dsin(crota2 * pi/180d0)
    rr = ra_pnt * pi/180d0
    dr = dec_pnt * pi/180d0  
    r0r = crval1 * pi/180d0
    d0r = crval2 * pi/180d0
    call spher2(rr, dr, r0r, d0r, dra, ddec)
    xr =   dra * cos_pa + ddec * sin_pa
    yr = - dra * sin_pa + ddec * cos_pa 
    xcen = xr * 180d0/pi / crdelt1 + crpix1
    ycen = yr * 180d0/pi / crdelt2 + crpix2

    if (xcen < 1.0 .or. xcen > npix(1) .or. &
        ycen < 1.0 .or. ycen > npix(2)) then
      xcen=min(max(xcen,1.0),real(npix(1)))
      ycen=min(max(ycen,1.0),real(npix(2)))
      call warning("WrongPointDir","Pointing direction is outside of image")
    end if

    call release(arr)
    call release(set)

  end subroutine co_image

  !---------------------------------------------------------------------
  subroutine co_bgima (fname, npix, xoff, yoff, inst, ipoint, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix 
    integer, dimension(2), save :: npix_save
    integer, dimension(3) :: ixoff, iyoff
    integer :: ix, iy, iix, iiy, i
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    real(kind=single) :: tot
    integer, dimension(:), pointer :: shape

    type(datasett) :: set
    type(arrayt) :: arr
    !!--JR
    logical :: withimbuff
    !!-- JR ------
    withimbuff = booleanParameter("withimagebuffersize")
    !!------------
    !!-- JR ------
    if (withimbuff) then
      imbuff = intParameter("imagebuffersize")
      withimbuff = .true.
      else
      imbuff = 640
      withimbuff = .false.
    endif
    !!------------

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)
    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    !! -- JR -----------------------------
!!!    if (withimbuff) then
!!!      imbuff = imbuff
!!!      else
    if (.not. withimbuff) then
      imbuff = max(npix(1),npix(2))
    end if
    !! -----------------------------------

    if (first) then
      npix_save=npix
      allocate (bg_image(nim,imbuff,imbuff))
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    tot=0.
    do iy = 1,npix(2)
      do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1) &
        !bg_image(ipoint,ix,iy) = 0.2
        bg_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = r4arr(iix,iiy)
        tot=tot+r4arr(iix,iiy)
      end do
    end do 
 
    if (tot<=1e-3) call error("EmptyBgMap","Background map is empty") 

    call release(arr)
    call release(set)
  
  end subroutine co_bgima

  !---------------------------------------------------------------------
  subroutine co_expima (fname, npix, xoff, yoff, inst, ipoint, &
                                            xcen, ycen, exp_center, nim)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint, nim 
    real(kind=single), intent(inout) :: xcen, ycen, exp_center
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix 

    integer, dimension(2), save :: npix_save
    integer, dimension(3) :: ixoff, iyoff    
    integer :: i, ix, iy, iix, iiy
    real(kind=single) :: tot, expmax
    logical :: first = .true.

    real(kind=single), dimension(:,:), pointer :: r4arr
    integer, dimension(:), pointer :: shape

    type(datasett) :: set
    type(arrayt) :: arr
    !!--JR
    logical :: withimbuff
    !!-- JR ------
    withimbuff = booleanParameter("withimagebuffersize")
    !!------------
    !!-- JR ------
    if (withimbuff) then
      imbuff = intParameter("imagebuffersize")
      withimbuff = .true.
      else
      imbuff = 640
      withimbuff = .false.
    endif
    !!------------

    set = dataset (fname, read)
    arr = array(set, "PRIMARY", read)

    if (hasAttribute(set,"XCEN")) then
      xcen = real32Attribute(set,"XCEN")
    else if (hasAttribute(arr,"XCEN")) then
      xcen = real32Attribute(arr,"XCEN")
    end if
    if (hasAttribute(set,"YCEN")) then
      ycen = real32Attribute(set,"YCEN")
    else if (hasAttribute(arr,"YCEN")) then
      ycen = real32Attribute(arr,"YCEN")
    end if
    exp_center=0.
    if (hasAttribute(set,"EXPOSURE")) then
      exp_center = real32Attribute(set,"EXPOSURE")
    else if (hasAttribute(arr,"EXPOSURE")) then
      exp_center = real32Attribute(arr,"EXPOSURE")
    end if 

    r4arr => real32array2data(arr)
    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    !! -- JR -----------------------------
    if (withimbuff) then
      imbuff = imbuff
      else
      imbuff = max(npix(1),npix(2))
    end if
    !! -----------------------------------

    if (first) then
!     print *,'allocate exp image'
      npix_save=npix
      allocate (exp_image(nim,imbuff,imbuff))
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    tot=0.
    expmax=0.
    do iy = 1,npix(2)
      do ix = 1,npix(1)
           iix=ix-1-ixoff(inst)
           iiy=iy-1-iyoff(inst)
           if (iix >= 0 .and. iix <=npix(1)-1 .and. &
               iiy >= 0 .and. iiy <=npix(2)-1) &
        exp_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = r4arr(iix,iiy)
        tot=tot+r4arr(iix,iiy)
        expmax=max(expmax,r4arr(iix,iiy))
      end do
    end do 
 
    if (tot<=1e-3) call error("EmptyExpMap","Exposure map is empty")
    if (exp_center<=0.) exp_center=expmax

    call release(arr)
    call release(set)

  end subroutine co_expima

  !---------------------------------------------------------------------
  subroutine co_mask (fname, npix, xoff, yoff, inst, ipoint)
  !---------------------------------------------------------------------

    character(len=1024), intent(in) :: fname
    integer, intent(in) :: inst, ipoint 
    real(kind=single), dimension(3), intent(in) :: xoff, yoff
    integer, dimension(2), intent(out) :: npix

    integer, dimension(2), save :: npix_save
    integer, dimension(3) :: ixoff, iyoff
    integer :: i, ix, iy, iix, iiy, atype, itot
    logical, save :: first = .true.

    integer, dimension(:,:), pointer :: i4arr
    integer, dimension(:), pointer :: shape
    real(kind=single), dimension(:,:), pointer :: r4arr

    type(datasett) :: set
    type(arrayt) :: arr
    !!--JR
    logical :: withimbuff
    !!-- JR ------
    withimbuff = booleanParameter("withimagebuffersize")
    !!------------
    !!-- JR ------
    if (withimbuff) then
      imbuff = intParameter("imagebuffersize")
      withimbuff = .true.
      else
      imbuff = 640
      withimbuff = .false.
    endif
    !!------------

    set = dataset (fname, read)

    !! The mask extension should be by default in an extension called
    !! MASK, if not then look in the PRIMARY array ! RDS
    if ( hasBlock( set, "MASK" ) ) then
        arr = array(set, "MASK", read)
    else
        arr = array(set, "PRIMARY", read)
    endif

    shape => dimensions(arr)
    npix(1) = shape(0)
    npix(2) = shape(1)

    !! -- JR -----------------------------
    if (withimbuff) then
      imbuff = imbuff
      else
      imbuff = max(npix(1),npix(2))
    end if
    !! -----------------------------------

    if (first) then
!     print *,'allocate mask image:', nim,npix
      npix_save=npix
      allocate (mask_image(nim,imbuff,imbuff))
      first=.false.
    end if

    if (npix_save(1) /= npix(1) .or. &
        npix_save(2) /= npix(2)) &
      call error("SizeMismatch:","Image "//trim(fname)//" has wrong size")

    do i=1,3
      ixoff(i)=nint(xoff(i))
      iyoff(i)=nint(yoff(i))
    end do

    itot=0
    atype = arrayDataType(arr)
    if (atype == Integer32) then
     i4arr => int32array2data(arr)
     do iy = 1,npix(2)
       do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1)  then
          if (i4arr(iix,iiy) == 0) then
            mask_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = .false.
            itot=itot+1
          else
            mask_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = .true.
          end if
        end if
       end do
     end do 
    end if
    if (atype == Real32) then
     r4arr => Real32array2data(arr)
     do iy = 1,npix(2)
       do ix = 1,npix(1)
        iix=ix-1-ixoff(inst)
        iiy=iy-1-iyoff(inst)
        if (iix >= 0 .and. iix <=npix(1)-1 .and. &
            iiy >= 0 .and. iiy <=npix(2)-1)  then
          if (r4arr(iix,iiy) < 1e-6) then
            mask_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = .false.
            itot=itot+1
          else
            mask_image(ipoint,ixb(ix,ipoint,1),iyb(iy,ipoint,1)) = .true.
          end if
        end if
       end do
     end do 
    end if

    if (itot==0) call error("EmptyMask","Detection mask is empty") 

    call release(arr)
    call release(set)

  end subroutine co_mask

end module edetect_io_box







